# -*- coding: utf-8 -*-
"""Purge monotonically named files in folders keeping range endpoints.

Implementation uses sha256 hashes for identity and assumes that
the natural order relates to the notion of fresher or better.
"""
import hashlib
import os
import typing

BUFFER_BYTES = 2 << 15


@typing.no_type_check
def list_dir(folder_path):
    """Access the dir and yield the local names inside."""
    return os.listdir(folder_path)


@typing.no_type_check
def elements_of_gen(folder_path):
    """Prefix names in folder path and yield sorted pairs of names and file paths."""
    for name in sorted(name for name in list_dir(folder_path)):
        yield name, os.path.join(folder_path, name)


@typing.no_type_check
def read_folder(folder_path, get_size=os.path.getsize):
    """Yield hash map of lists with name, byte size pairs of sorted by name (hint: timestamp)."""
    hash_map = {}
    for name, file_path in elements_of_gen(folder_path):
        with open(file_path, 'rb') as in_file:
            sha256_hash = hashlib.sha256()
            for byte_block in iter(lambda in_f=in_file: in_f.read(BUFFER_BYTES), b''):
                sha256_hash.update(byte_block)
            hash_map.setdefault(sha256_hash.hexdigest(), []).append((name, get_size(file_path)))

    return hash_map


@typing.no_type_check
def triage_hashes(hash_map):
    """Triage hash map in pair of names to keep and to remove in that order.

    Three cases:

    0. size zero regardless of hash => remove
    1. unique hash => keep
    2. hash matching two entries => keep both
    3. hash with more than two entries => keep first and last, rest remove
    """
    keep, remove = [], []
    for info in hash_map.values():
        if info[0][1] == 0:
            remove.extend(name for name, _ in info)
        else:
            if len(info) == 1:
                keep.extend(name for name, _ in info)
            else:
                first, last = info[0][0], info[-1][0]
                keep.extend([first, last])
                remove.extend(name for name, _ in info[1:-1])

    return keep, remove



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
"""Purge monotonically named files in folders keeping range endpoints.

Implementation uses sha256 hashes for identity and assumes that
the natural order relates to the notion of fresher or better.
"""
import datetime as dti
import os
import sys
import typing

from puristaa.puristaa import prefix_compression  # type: ignore

from puhdistusalue.puhdistusalue import read_folder, triage_hashes

DEBUG = os.getenv('PURGE_RANGE_DEBUG')


@typing.no_type_check
def humanize_mass(total_less_bytes: int):
    """DRY"""
    if total_less_bytes >= 1e9:
        return f'{round(total_less_bytes / 1024 / 1024 / 1024, 3) :.3f}', 'total gigabytes'
    elif total_less_bytes >= 1e6:
        return f'{round(total_less_bytes / 1024 / 1024, 3) :.3f}', 'total megabytes'
    elif total_less_bytes >= 1e3:
        return f'{round(total_less_bytes / 1024, 3) :.3f}', 'total kilobytes'

    return f'{total_less_bytes :d}', 'total bytes'


@typing.no_type_check
def humanize_duration(duration_seconds: float):
    """DRY"""
    if duration_seconds >= 3600:
        return f'{round(duration_seconds / 60 / 60, 3) :.3f}', 'hours'
    elif duration_seconds >= 60:
        return f'{round(duration_seconds / 60, 3) :.3f}', 'minutes'
    elif duration_seconds >= 1:
        return f'{round(duration_seconds, 3) :.3f}', 'seconds'

    return f'{round(duration_seconds * 1e3, 3) :.3f}', 'millis'


# pylint: disable=expression-not-assigned
@typing.no_type_check
def main(argv=None):
    """Process the files separately per folder."""
    start_time = dti.datetime.utcnow()
    argv = sys.argv[1:] if argv is None else argv
    verbose = True if '-v' in argv or '--verbose' in argv else False
    human = True if '-H' in argv or '--human' in argv else False
    folder_paths = [entry for entry in argv if entry.strip() and entry not in ('-H', '--human', '-v', '--verbose')]
    total_removed, total_less_bytes = 0, 0
    for a_path in folder_paths:
        hash_map = {}
        try:
            hash_map = read_folder(a_path)
        except FileNotFoundError as err:
            print(f'WARNING: Skipping non-existing path ({a_path}) -> "{err}"')
        if hash_map:
            keep_these, remove_those = triage_hashes(hash_map)
            for this in keep_these:
                DEBUG and print(f'KEEP file {this}')
            folder_removed, folder_less_bytes = 0, 0
            for that in remove_those:
                DEBUG and print(f'REMOVE file {that}')
                target = os.path.join(a_path, that)
                folder_less_bytes += os.path.getsize(target)
                os.remove(target)
                folder_removed += 1

            if verbose:
                print(
                    f'removed {folder_removed} redundant objects or {folder_less_bytes}'
                    f' combined bytes from folder at {a_path}'
                )
            total_less_bytes += folder_less_bytes
            total_removed += folder_removed

    prefix, rel_paths = prefix_compression(folder_paths, policy=lambda x: x == '/')
    if len(rel_paths) > 5:
        folders_disp = f"{prefix}[{', '.join(rel_paths[:3])}, ... {rel_paths[-1]}]"
    else:
        folders_disp = f'{folder_paths}' if folder_paths else '[<EMPTY>]'

    duration_seconds = (dti.datetime.utcnow() - start_time).total_seconds()
    if human:
        m_quantity, m_unit = humanize_mass(total_less_bytes)
        d_quantity, d_unit = humanize_duration(duration_seconds)
    else:
        m_quantity, m_unit = f'{total_less_bytes :d}', 'total bytes'
        d_quantity, d_unit = f'{round(duration_seconds, 3) :.3f}', 'seconds'

    print(
        f'removed {total_removed} total redundant objects or'
        f' {m_quantity} {m_unit} from folders at {folders_disp}'
        f' in {d_quantity} {d_unit}'
    )



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import pytest  # type: ignore
from puristaa.puristaa import prefix_compression  # type: ignore

import puhdistusalue.puhdistusalue as pr


def test_triage_hashes_ok_empty_hash_map():
    assert pr.triage_hashes({}) == ([], [])


def test_prefix_compression_ok_string():
    texts = 'imension is implicit'
    assert prefix_compression(texts) == (texts, [''])


def test_purge_range_ok_same_chars_in_string():
    texts = 'a a a a'
    assert prefix_compression(texts) == (texts, [''])


def test_purge_range_ok_sequence_string():
    assert prefix_compression(['aa']) == ('aa', [''])


def test_purge_range_ok_strings():
    assert prefix_compression(['aa', 'ab']) == ('a', ['a', 'b'])


def test_purge_range_ok_disjoint_strings():
    texts = ['a', 'b']
    assert prefix_compression(texts) == ('', texts)


def test_purge_range_ok_empty():
    texts = []
    assert prefix_compression(texts) == ('', texts)


def test_purge_range_nok_ints():
    message = r"'int' object is not iterable"
    with pytest.raises(TypeError, match=message):
        prefix_compression([1, 2, 3])


def test_purge_range_nok_floats():
    message = r"'float' object is not iterable"
    with pytest.raises(TypeError, match=message):
        prefix_compression([0.123, 3.1415])


def test_prefix_compression_documentation_ok_example():
    sequence = ['bar/baz', 'bar/bazaar']
    expect = ('bar/', ['baz', 'bazaar'])
    assert prefix_compression(sequence, policy=lambda x: x == '/') == expect


def test_prefix_compression_documentation_ok_no_policy_example():
    sequence = ['bar/baz', 'bar/bazaar']
    expect = ('bar/baz', ['', 'aar'])
    assert prefix_compression(sequence, policy=None) == expect


def test_prefix_compression_documentation_ok_tuple_no_policy_example():
    sequence = ('bar/baz', 'bar/bazaar')
    expect = ('bar/baz', ['', 'aar'])
    assert prefix_compression(sequence, policy=None) == expect


def test_prefix_compression_documentation_nok_set_no_policy_example():
    sequence = {'bar/baz', 'bar/bazaar'}
    message = r"'set' object is not subscriptable"
    with pytest.raises(TypeError, match=message):
        prefix_compression(sequence, policy=None)


def test_prefix_compression_documentation_ok_dict_no_policy_example():
    mapping = {0: 'bar/baz', 1: 'bar/bazaar'}
    expect = ('', [{0: 'bar/baz', 1: 'bar/bazaar'}])
    assert prefix_compression(mapping, policy=None) == expect


def test_prefix_compression_documentation_nok_class_instance_no_policy_example():
    class Foo:
        pass

    an_object = Foo()
    message = r"'Foo' object is not iterable"
    with pytest.raises(TypeError, match=message):
        prefix_compression(an_object, policy=None)


def test_triage_hashes_ok_mixed_input():
    data = {
        'no': [('matter', 4), ('remove', 4), ('what', 4)],
        'we': [('keep', 1)],
        'maybe': [('empty', 0)],
    }
    keep, remove = pr.triage_hashes(data)
    assert keep == ['matter', 'what', 'keep']
    assert remove == ['remove', 'empty']


def test_triage_hashes_ok_mixed_input_no_reorder():
    data = {
        'a': [('z', 4), ('a', 4), ('y', 4)],
        'b': [('x', 1), ('y', 2), ('x', 3)],
        'c': [('_', 0)],
    }
    keep, remove = pr.triage_hashes(data)
    assert keep == ['z', 'y', 'x', 'x']
    assert remove == ['a', 'y', '_']



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import pathlib

import pytest  # type: ignore

import puhdistusalue.cli as cli


def test_main_nok_empty():
    assert cli.main([]) is None


def test_main_nok_int():
    message = r"argument of type 'int' is not iterable"
    with pytest.raises(TypeError, match=message):
        cli.main(42)


def test_main_nok_ints():
    sequence_of_ints = [1, 2, 3]
    nef = str(sequence_of_ints[0])
    assert pathlib.Path(nef).is_dir() is False, f'Unexpected folder {nef} exists which breaks this test'
    message = r"'int' object has no attribute 'strip'"
    with pytest.raises(AttributeError, match=message):
        cli.main(sequence_of_ints)


def test_main_nok_non_existing_folder():
    nef = non_existing_folder_path = 'folder_does_not_exist'
    assert pathlib.Path(nef).is_dir() is False, f'Unexpected folder {nef} exists which breaks this test'
    assert cli.main([non_existing_folder_path]) is None


def test_main_nok_non_existing_folder_verbose():
    nef = non_existing_folder_path = 'folder_does_not_exist'
    assert pathlib.Path(nef).is_dir() is False, f'Unexpected folder {nef} exists which breaks this test'
    assert cli.main([non_existing_folder_path, '-v']) is None


def test_main_ok_distinct_timestamps_folder(capsys):
    dist_ts_folder = str(pathlib.Path('tests', 'fixtures', 'timestamps', 'all_distinct'))
    cli.main([dist_ts_folder])
    out, err = capsys.readouterr()
    assert 'removed 0 total redundant objects or 0 total bytes' in out
    assert dist_ts_folder in out
    assert not err
